# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import zlib
import gzip

from libcloud.utils.py3 import PY3
from libcloud.utils.py3 import StringIO


__all__ = [
    'decompress_data'
]


def decompress_data(compression_type, data):
    if compression_type == 'zlib':
        return zlib.decompress(data)
    elif compression_type == 'gzip':
        # TODO: Should export BytesIO as StringIO in libcloud.utils.py3
        if PY3:
            from io import BytesIO
            cls = BytesIO
        else:
            cls = StringIO

        return gzip.GzipFile(fileobj=cls(data)).read()
    else:
        raise Exception('Invalid or onsupported compression type: %s' %
                        (compression_type))
